#include "CSolver.hpp"
#include "RBody.hpp"
#include "CDetector.hpp"
#include "ODESolver.hpp"
#include "main.h"
#include <math.h>
#include <cstdio>

#define FRICTION 0.1
#define STATIC_FRICTION 0.12

void CSolver::SolveAllCollisions(std::vector<Collision*>* collisions, double delta_time)
{
  while (collisions->size() > 0)
  {
    Collision* col = collisions->back();
    SolveCollision(col, delta_time);
    delete col;
    collisions->pop_back();
  }
}

void SolveOnePoint(Vector3 p, Vector3 n, RBody* A, RBody* B, bool default_ela = true, double elasticity = 0)
{
  double e = A->elasticity * B->elasticity;
  if (!default_ela)
    e = elasticity;
  
  Vector3 pa = p - A->pos;
  Vector3 pb = p - B->pos; 
  Vector3 vpa = A->vel + (A->ang_vel ^ pa);
  Vector3 vpb = B->vel + (B->ang_vel ^ pb);
  Vector3 rel_vel_v = vpa - vpb;
  double rel_vel = n*rel_vel_v;
  
  double inv_m_a = 1/A->mass;
  double inv_m_b = 1/B->mass;
  Matrix3x3 inv_I_a = A->InvInertiaTensor();
  Matrix3x3 inv_I_b = B->InvInertiaTensor();
  
  if (A->is_static || A->is_frozen)
  {
    inv_m_a = 0;
    inv_I_a = Matrix3x3();
  }
  
  if (B->is_static || B->is_frozen)
  {
    inv_m_b = 0;
    inv_I_b = Matrix3x3();
  }
  
  if (inv_m_a == 0 && inv_m_b == 0) return;
  
  double a = -(1+e)*rel_vel;
  double b = n*((inv_I_a*(pa^n))^pa);
  double c = n*((inv_I_b*(pb^n))^pb);
  
  double j = a / (b + c + inv_m_a + inv_m_b);
  Vector3 J = n*j;
  
  if (!(A->is_static) && !(A->is_frozen))
  {
    A->vel += J*inv_m_a;
    A->ang_vel += inv_I_a*(pa^J);
  }
  if (!(B->is_static) && !(B->is_frozen))
  {
    B->vel -= J*inv_m_b;
    B->ang_vel -= inv_I_b*(pb^J);
  }
  // friction
  
  Vector3 t = n^(n^rel_vel_v);
  
  if (t.Length() < 0.01) // if n and rel_vel_v are perpendicular we apply no friction
    return;
  
  t = t.Normalized();
  double rel_vel_t = t*rel_vel_v;
  if (fabs(rel_vel_t) < 0.0001) return;
  
  
  double a_f = -rel_vel_t;
  double b_f = t*((inv_I_a*(pa^t))^pa);
  double c_f = t*((inv_I_b*(pb^t))^pb);
  
  double j_f_max = a_f / (b_f + c_f + inv_m_a + inv_m_b);
  double j_f = fmin(j_f_max, FRICTION*j);
  
  Vector3 J_f;
  if (j_f < STATIC_FRICTION * j)
    J_f = t * j_f_max;
  else
    J_f = t * j_f;
  
  if (!(A->is_static) && !(A->is_frozen))
  {
    A->vel += J_f*inv_m_a;
    A->ang_vel += inv_I_a*(pa^J_f);
  }
  if (!(B->is_static) && !(B->is_frozen))
  {
    B->vel -= J_f*inv_m_b;
    B->ang_vel -= inv_I_b*(pb^J_f);
  }
}

/*void CSolver::UpdateVelocities(Collision* collision)
{
  std::vector<CollisionPoint>::iterator it;
  for (it = collision->points.begin(); it != collision->points.end(); ++it)
  {
    it->rel_vel = CDetector::RelativeVelocity(it->p, it->n, collision->A, collision->B);
  }
}*/

/*void RemoveSeparating(Collision* collision)
{
  std::vector<CollisionPoint>::iterator it;
  for (it = collision->points.begin(); it != collision->points.end();)
  {
    if (it->rel_vel > 0)
      it = collision->points.erase(it);
    else
      it++;
  }
}

void RemoveNonpenetrating(Collision* collision)
{
  std::vector<CollisionPoint>::iterator it;
  for (it = collision->points.begin(); it != collision->points.end();)
  {
    if (!CDetector::IsPointInside(it->p, it->normaler))
      it = collision->points.erase(it);
    else
      it++;
  }
}*/
  
void CSolver::SolveCollision(Collision* collision, double delta_time)
{
  RBody* A = collision->bodies[0];
  RBody* B = collision->bodies[1];
  
  float max_depth = -1;
  Vector3 max_point;
  Vector3 max_normal;
  
  for (int body = 0; body < 2; body++)
  {
    for (unsigned int i = 0; i < collision->points[body].size(); i++)
    {
      CollisionPoint cp = collision->points[body][i];
      PointEvaluation pe = CDetector::EvaluatePoint(cp, collision->bodies[body], collision->bodies[(body+1)%2]);
      float rel_vel = CDetector::RelativeVelocity(pe.p, pe.n, collision->bodies[body],  collision->bodies[(body+1)%2]);
      if (pe.depth > max_depth+0.0001 && rel_vel < -0.001)
      {
        max_depth = pe.depth;
        max_point = pe.p;
        max_normal = pe.n;
        if (body == 1)
          max_normal = max_normal * -1;
      }
    }
  }

  if (max_depth < 0)
    return;

  SolveOnePoint(max_point, max_normal, A, B);
}

void CSolver::SolveContact(Collision* collision, double delta_time, double elasticity)
{
  RBody* A = collision->bodies[0];
  RBody* B = collision->bodies[1];

  //printf("solving contact\n");

  for (int iter = 0; iter < 1; iter++)
  {
    float max_depth = -1;
    Vector3 max_point;
    Vector3 max_normal;
    
    float depth_sum = 0;
    Vector3 point_sum;
    
    for (int body = 0; body < 2; body++)
    {
      for (unsigned int i = 0; i < collision->points[body].size(); i++)
      {
        CollisionPoint cp = collision->points[body][i];
        PointEvaluation pe = CDetector::EvaluatePoint(cp, collision->bodies[body], collision->bodies[(body+1)%2]);
        float rel_vel = CDetector::RelativeVelocity(pe.p, pe.n, collision->bodies[body], collision->bodies[(body+1)%2]);
        if (pe.depth > max_depth+0.0001 && rel_vel < -0.0001)
        {
          max_depth = pe.depth;
          max_point = pe.p;
          max_normal = pe.n;
          if (body == 1)
            max_normal = max_normal * -1;
        }
        point_sum += pe.p*pe.depth;
        depth_sum += pe.depth;
      }
    }
    
    max_point = point_sum / depth_sum;

    if (max_depth < 0)
      break;
    global::DumpPoints(max_point, max_depth);
    SolveOnePoint(max_point, max_normal, A, B, false, elasticity);
  }
}

/*void CSolver::SolvePrecise(Collision* collision, bool default_ela, double elasticity)
{
  RBody* A = collision->A;
  RBody* B = collision->B;
  UpdateVelocities(collision);
  
  while (collision->points.size() > 0)
  {    
    std::vector<CollisionPoint>::iterator it = collision->points.begin();
    std::vector<CollisionPoint>::iterator max_pen = it;
    for (; it != collision->points.end(); it++)
    {
      if (it->depth > max_pen->depth)
        max_pen = it;
    }
    SolveOnePoint(*max_pen, A, B, default_ela, elasticity);
    collision->points.erase(max_pen);
    UpdateVelocities(collision);
    
    
    int non_separating_count = 0;
    for (it = collision->points.begin(); it != collision->points.end(); it++)
    {
      if (it->rel_vel > 0) non_separating_count++;
    }
    if (non_separating_count == 0)
      break;
  }
}*/
