#include "Network.hpp"
#include <enet/enet.h>
#include <cstdio>

#define PORT 1234

namespace Network
{
  ENetAddress address; // address of the host
  ENetHost* host;
  ENetPeer* peer;
  bool is_server;
  
  std::vector<RBody*>* bodies;
  std::vector<Player>* players;
  
  static union
	{
		char bytes[sizeof(double)];
		double val;
	} bdconvert;
  
  static union
	{
		char bytes[sizeof(unsigned int)];
		unsigned int val;
	} biconvert;
}

using namespace Network;

double ReadDouble(enet_uint8* array, int* pos)
{
	int position = *pos;
  for (unsigned int i = 0; i < sizeof(double); i++)
    bdconvert.bytes[i] = array[position++];
	
  double ret = bdconvert.val;
	*pos = position;
	return ret;
} 

void WriteDouble(enet_uint8* array, int* pos, double x)
{
	int position = *pos;
	bdconvert.val = x;
	
  for (unsigned int i = 0; i < sizeof(double); i++)
    array[position++] = bdconvert.bytes[i];
    
	*pos = position;
}

unsigned int ReadInt(enet_uint8* array, int* pos)
{
	int position = *pos;
  for (unsigned int i = 0; i < sizeof(unsigned int); i++)
    biconvert.bytes[i] = array[position++];
	
  unsigned int ret = biconvert.val;
	*pos = position;
	return ret;
} 

void WriteInt(enet_uint8* array, int* pos, unsigned int x)
{
	int position = *pos;
	biconvert.val = x;
	
  for (unsigned int i = 0; i < sizeof(unsigned int); i++)
    array[position++] = biconvert.bytes[i];
    
	*pos = position;
}

Vector3 ReadVector(enet_uint8* array, int* pos)
{
  Vector3 ret;
  ret[0] = ReadDouble(array, pos);
  ret[1] = ReadDouble(array, pos);
  ret[2] = ReadDouble(array, pos);
  return ret;
}

void WriteVector(enet_uint8* array, int* pos, Vector3 vec)
{
  WriteDouble(array, pos, vec.x());
  WriteDouble(array, pos, vec.y());
  WriteDouble(array, pos, vec.z());
}

Quaternion ReadQuaternion(enet_uint8* array, int* pos)
{
  Quaternion ret;
  ret.s = ReadDouble(array, pos);
  ret.v = ReadVector(array, pos);
  return ret;
}

void WriteQuaternion(enet_uint8* array, int* pos, Quaternion q)
{
  WriteDouble(array, pos, q.s);
  WriteVector(array, pos, q.v);
}

bool Network::IsConnected()
{
  if (host != NULL && peer != NULL)
    return true;
  else
    return false;
}

void Network::Connect(std::string address_str, std::vector<RBody*>* bodies_ptr, std::vector<Player>* players_ptr)
{
  bodies = bodies_ptr;
  players = players_ptr;
  is_server = false;
  
  if (enet_initialize() != 0)
  {
    printf("An error occurred while initializing ENet.\n");
    return;
  }
  
  host = enet_host_create(NULL, 1, 2, 57600/8, 14400/8);
  if (host == NULL)
  {
    printf("An error occurred while trying to create an ENet client host.\n");
    exit(EXIT_FAILURE);
  }
  
  ENetEvent event;

  enet_address_set_host(&address, address_str.c_str());
  address.port = 1234;
  peer = enet_host_connect(host, &address, 2, 0);

  if (peer == NULL)
  {
    printf("No available peers for initiating an ENet connection.\n");
    exit (EXIT_FAILURE);
  }
  
  if (enet_host_service(host, &event, 5000) > 0 && event.type == ENET_EVENT_TYPE_CONNECT)
  {
    printf("Connection to %s:%d succeeded.\n", address_str.c_str(), PORT);
  }
  else
  {
    enet_peer_reset(peer);
    printf("Connection to %s:%d failed.\n", address_str.c_str(), PORT);
  }
}

void Network::CreateHost(std::vector<RBody*>* bodies_ptr, std::vector<Player>* players_ptr)
{
  bodies = bodies_ptr;
  players = players_ptr;
  is_server = true;
  
  if (enet_initialize() != 0)
  {
    printf("An error occurred while initializing ENet.\n");
    return;
  }
  
  address.host = ENET_HOST_ANY;
  address.port = PORT;
  
  host = enet_host_create(&address, 1, 2, 0, 0);
  
  if (host == NULL)
  {
    printf("An error occurred while trying to create an ENet server host.\n");
    exit(EXIT_FAILURE);
  }
}

void Network::CloseHost()
{
  enet_host_destroy(host);
  enet_deinitialize();
}

void Network::CloseClient()
{
  enet_host_destroy(host);
  enet_deinitialize();
}

void Network::SendInput(Input in, unsigned int simulation_time, unsigned int controlled_idx)
{
  if (peer == NULL)
    return;
    
  enet_uint8 data[16];  
	int pos = 0;
  
  data[pos++] = MSG_SEND_INPUT;
  WriteInt(data, &pos, simulation_time);
  data[pos++] = in;
  WriteInt(data, &pos, controlled_idx);
  
  ENetPacket* packet = enet_packet_create(data, pos, ENET_PACKET_FLAG_RELIABLE);
  enet_peer_send(peer, 0, packet);
  enet_host_flush(host);
}

void SendSyncBodies(unsigned int simulation_time, unsigned int new_id)
{
  printf("Sending sync at %i\n", simulation_time);
  
  enet_uint8 data[1024];  
	int pos = 0;
  
  data[pos++] = MSG_SYNC_BODIES;
  WriteInt(data, &pos, simulation_time);
  data[pos++] = bodies->size();
  
  for (unsigned int i = 0; i < bodies->size(); i++)
  {
    RBody* body = (*bodies)[i];
    
    WriteDouble(data, &pos, body->density);
    WriteDouble(data, &pos, body->width);
    WriteDouble(data, &pos, body->height);
    WriteDouble(data, &pos, body->depth);
    
    WriteVector(data, &pos, body->vel);
    WriteVector(data, &pos, body->pos);
    WriteVector(data, &pos, body->ang_vel);
    WriteQuaternion(data, &pos, body->GetAngPos());
    
    body->vel.Print();
    body->pos.Print();
    body->ang_vel.Print();
    body->GetAngPos().Print();
    
    data[pos++] = body->is_static;
  }
  
  WriteInt(data, &pos, new_id);
  
  ENetPacket* packet = enet_packet_create(data, pos, ENET_PACKET_FLAG_RELIABLE);
  enet_peer_send(peer, 0, packet);
  enet_host_flush(host);
}

Command ReadPacket(ENetPacket* packet)
{
  Command command;
  
  if (packet->data[0] == MSG_SYNC_BODIES)
  {    
    int pos = 1;
    int last_id = 0;
    
    command.type = MSG_SYNC_BODIES;
    command.simulation_time = ReadInt(packet->data, &pos);
    
    char body_num = packet->data[pos++];
    
    printf("Received sync at %i, %i bodies\n", command.simulation_time, body_num);
    
    for (int i = 0; i < body_num; i++)
    {
      double d = ReadDouble(packet->data, &pos);
      double x = ReadDouble(packet->data, &pos);
      double y = ReadDouble(packet->data, &pos);
      double z = ReadDouble(packet->data, &pos);
      
      RBody* body = new RBody(d, x, y, z, "body");
      body->id = last_id;
      last_id++;
      body->vel = ReadVector(packet->data, &pos);
      body->pos = ReadVector(packet->data, &pos);
      body->ang_vel = ReadVector(packet->data, &pos);
      body->SetAngPos(ReadQuaternion(packet->data, &pos));
      
      body->is_static = packet->data[pos++];
      
      body->vel.Print();
      body->pos.Print();
      body->ang_vel.Print();
      body->GetAngPos().Print();
      
      bodies->push_back(body);
    }
    command.id = ReadInt(packet->data, &pos);
  }
  else if (packet->data[0] == MSG_SEND_INPUT)
  {
    printf("Received new input.\n");
    
    int pos = 1;
    
    command.type = MSG_SEND_INPUT;
    command.simulation_time = ReadInt(packet->data, &pos);
    command.input = (Input)packet->data[pos++];
    command.id = ReadInt(packet->data, &pos);
  }
  
  return command;
}

Command Network::HostLoop(unsigned int simulation_time, double* speed)
{
  ENetEvent event;
  Command command;
  unsigned int new_player_id;
  command.type = MSG_NONE;
  
  if (enet_host_service(host, &event, 0) == 0)
    return command;

  switch (event.type)
  {
    case ENET_EVENT_TYPE_CONNECT:
      printf("A new client connected from %x:%u.\n",
        event.peer->address.host,
        event.peer->address.port);
      peer = event.peer;
      new_player_id = players->size();
      SendSyncBodies(simulation_time, new_player_id);
      players->push_back(Player(new_player_id, (*bodies)[new_player_id]));
      break;
    case ENET_EVENT_TYPE_RECEIVE:
      command = ReadPacket(event.packet);
      enet_packet_destroy (event.packet);
      break;
    case ENET_EVENT_TYPE_DISCONNECT:
      printf("%s disconnected.\n", (char*)(event.peer->data));
      event.peer->data = NULL;
      peer = NULL;
      break;
    case ENET_EVENT_TYPE_NONE:
      break;
  }
  
  return command;
}

Command Network::ClientLoop(double* speed)
{
  ENetEvent event;
  Command command;
  command.type = MSG_NONE;
  
  if (enet_host_service(host, &event, 0) != 0)
  {
    switch (event.type)
    {
      case ENET_EVENT_TYPE_CONNECT:
        break;
      case ENET_EVENT_TYPE_RECEIVE:
        command = ReadPacket(event.packet);
        enet_packet_destroy (event.packet);
        break;
      case ENET_EVENT_TYPE_DISCONNECT:
        break;
      case ENET_EVENT_TYPE_NONE:
        break;
    }
  }
  
  return command;
}
