#include "RBody.hpp"

#define COL_POINT_BUFFER
#define STATE_LIST_LEN 50

// for now all rigid bodies are boxes
RBody::RBody(double dens, double x_size, double y_size, double z_size, std::string str)
{
  name = str;
  density = dens;
  width = x_size;
  height = y_size;
  depth = z_size;
  
  is_static = false;  
  is_frozen = false;
  mass = x_size*y_size*z_size;
  pos = Vector3();
  vel = Vector3();
  ang_vel = Vector3();
  ang_pos = Quaternion(1,0,0,0);
  ang_pos = ang_pos.Normalize();
  elasticity = 0.5f;
  recent_cols_last = 0;
  recent_cols_size = 0;
  
  local_inertia = Matrix3x3();
  local_inertia(0,0) = (y_size*y_size + z_size*z_size)/12.0f * mass * 2.0f;
  local_inertia(1,1) = (x_size*x_size + z_size*z_size)/12.0f * mass * 2.0f;
  local_inertia(2,2) = (y_size*y_size + x_size*x_size)/12.0f * mass * 2.0f;
  
  inv_local_inertia = local_inertia.Inverse();
  
  colour = Vector3(0.5,0.5,1);
  
  for (int z = -1; z <= 1; z += 2)
  {
    Face face;
    face.vertices.push_back(Vector3(-x_size/2, -y_size/2, z*z_size/2));
    face.vertices.push_back(Vector3(-x_size/2, y_size/2, z*z_size/2));
    face.vertices.push_back(Vector3(x_size/2, y_size/2, z*z_size/2));
    face.vertices.push_back(Vector3(x_size/2, -y_size/2, z*z_size/2));
    face.n = Vector3(0,0,z);
    faces.push_back(face);
    
    Edge edge;
    edge.a = Vector3(-x_size/2, -y_size/2, z*z_size/2);
    edge.b = Vector3(-x_size/2, y_size/2, z*z_size/2);
    edge.dir = (edge.a - edge.b).Normalized();
    edges.push_back(edge);
    
    edge.a = Vector3(x_size/2, -y_size/2, z*z_size/2);
    edge.b = Vector3(x_size/2, y_size/2, z*z_size/2);
    edge.dir = (edge.a - edge.b).Normalized();
    edges.push_back(edge);
  }
  for (int y = -1; y <= 1; y += 2)
  {
    Face face;
    face.vertices.push_back(Vector3(-x_size/2, y*y_size/2, -z_size/2));
    face.vertices.push_back(Vector3(-x_size/2, y*y_size/2, z_size/2));
    face.vertices.push_back(Vector3(x_size/2, y*y_size/2, z_size/2));
    face.vertices.push_back(Vector3(x_size/2, y*y_size/2, -z_size/2));
    face.n = Vector3(0,y,0);
    faces.push_back(face);
    
    Edge edge;
    edge.a = Vector3(x_size/2, y*y_size/2, -z_size/2);
    edge.b = Vector3(-x_size/2, y*y_size/2, -z_size/2);
    edge.dir = (edge.a - edge.b).Normalized();
    edges.push_back(edge);
    
    edge.a = Vector3(x_size/2, y*y_size/2, z_size/2);
    edge.b = Vector3(-x_size/2, y*y_size/2, z_size/2);
    edge.dir = (edge.a - edge.b).Normalized();
    edges.push_back(edge);
  }
  for (int x = -1; x <= 1; x += 2)
  {
    Face face;
    face.vertices.push_back(Vector3(x*x_size/2, -y_size/2, -z_size/2));
    face.vertices.push_back(Vector3(x*x_size/2, -y_size/2, z_size/2));
    face.vertices.push_back(Vector3(x*x_size/2, y_size/2, z_size/2));
    face.vertices.push_back(Vector3(x*x_size/2, y_size/2, -z_size/2));
    face.n = Vector3(x,0,0);
    faces.push_back(face);
    
    Edge edge;
    edge.a = Vector3(x*x_size/2, -y_size/2, z_size/2);
    edge.b = Vector3(x*x_size/2, -y_size/2, -z_size/2);
    edge.dir = (edge.a - edge.b).Normalized();
    edges.push_back(edge);
    
    edge.a = Vector3(x*x_size/2, y_size/2, z_size/2);
    edge.b = Vector3(x*x_size/2, y_size/2, -z_size/2);
    edge.dir = (edge.a - edge.b).Normalized();
    edges.push_back(edge);
  }
  
  for (int x = -1; x <= 1; x+=2)
    for (int y = -1; y <= 1; y+=2)
      for (int z = -1; z <= 1; z+=2)
        vertices.push_back(Vector3(x*x_size/2, y*y_size/2, z*z_size/2));
        
  bounding_radius = Vector3(x_size/2, y_size/2, z_size/2).Length();
  rot_matrix_valid = false;
}

Vector3 RBody::Forces(RBState state) const
{
  if (is_static || is_frozen)
    return Vector3(0,0,0);
  else
    return forces;
}

Vector3 RBody::Torque(RBState state) const
{
  if (is_static || is_frozen)
    return Vector3(0,0,0);
  else
    return torque;
}

void RBody::ResetForceTorque()
{
  forces = torque = Vector3(0,0,0);
}

RBState RBody::State() const
{
  RBState ret;
  ret.vel = vel;
  ret.pos = pos;
  ret.ang_vel = ang_vel;
  ret.ang_pos = ang_pos;
  return ret;
}

void RBody::ApplyState(RBState state)
{
  ApplyPos(state);
  ApplyVel(state);
}

void RBody::ApplyPos(RBState state)
{
  pos = state.pos;
  ang_pos = state.ang_pos;
}

void RBody::ApplyVel(RBState state)
{
  vel = state.vel;
  ang_vel = state.ang_vel;
}

Matrix3x3 RBody::RotationMatrix()
{
  if (rot_matrix_valid)
    return rot_matrix;
    
  Matrix3x3 ret;
  double s = ang_pos.s;
  Vector3 v = ang_pos.v;
  ret(0,0) = 1 - 2*v.y()*v.y() - 2*v.z()*v.z();
  ret(0,1) = 2*v.x()*v.y() - 2*s*v.z();
  ret(0,2) = 2*v.x()*v.z() + 2*s*v.y();
  ret(1,0) = 2*v.x()*v.y() + 2*s*v.z();
  ret(1,1) = 1 - 2*v.x()*v.x() - 2*v.z()*v.z();
  ret(1,2) = 2*v.y()*v.z() - 2*s*v.x();
  ret(2,0) = 2*v.x()*v.z() - 2*s*v.y();
  ret(2,1) = 2*v.y()*v.z() + 2*s*v.x();
  ret(2,2) = 1 - 2*v.x()*v.x() - 2*v.y()*v.y();
  
  rot_matrix = ret;
  rot_matrix_valid = true;
  
  return ret;
}

Matrix3x3 RBody::RotationMatrix(Quaternion q)
{
  Matrix3x3 ret;
  double s = q.s;
  Vector3 v = q.v;
  ret(0,0) = 1 - 2*v.y()*v.y() - 2*v.z()*v.z();
  ret(0,1) = 2*v.x()*v.y() - 2*s*v.z();
  ret(0,2) = 2*v.x()*v.z() + 2*s*v.y();
  ret(1,0) = 2*v.x()*v.y() + 2*s*v.z();
  ret(1,1) = 1 - 2*v.x()*v.x() - 2*v.z()*v.z();
  ret(1,2) = 2*v.y()*v.z() - 2*s*v.x();
  ret(2,0) = 2*v.x()*v.z() - 2*s*v.y();
  ret(2,1) = 2*v.y()*v.z() + 2*s*v.x();
  ret(2,2) = 1 - 2*v.x()*v.x() - 2*v.y()*v.y();
  
  return ret;
}

RBState RBody::StateDerivative(RBState s) const
{
  RBState derivative;
  Matrix3x3 rot_matrix = RBody::RotationMatrix(s.ang_pos);
  Matrix3x3 inv_inertia = rot_matrix * inv_local_inertia * rot_matrix.Transpose();
  
  derivative.pos = s.vel;
  derivative.vel = Forces(s) / mass;
  derivative.ang_pos = (Quaternion(0, s.ang_vel) * s.ang_pos) * 0.5f;
  derivative.ang_vel = inv_inertia * Torque (s);

  return derivative;
}

Face RBody::GetFace(int i)
{
  return faces[i];
}

int RBody::FaceNum()
{
  return faces.size();
}

Vector3 RBody::PointLocalToGlobal(Vector3 v)
{
  return RotationMatrix()*v + pos;
}

Vector3 RBody::VectorLocalToGlobal(Vector3 v)
{
  return RotationMatrix()*v;
}

Vector3 RBody::PointGlobalToLocal(Vector3 v)
{
  return RotationMatrix().Inverse()*(v - pos);
}

Vector3 RBody::VectorGlobalToLocal(Vector3 v)
{
  return RotationMatrix().Inverse()*v;
}

Matrix3x3 RBody::InertiaTensor()
{
  Matrix3x3 rot_mat = RotationMatrix();
  return rot_mat*local_inertia*(rot_mat.Transpose());
}

Matrix3x3 RBody::InvInertiaTensor()
{
  Matrix3x3 rot_mat = RotationMatrix();
  return rot_mat*inv_local_inertia*(rot_mat.Transpose());
}

Quaternion RBody::GetAngPos() const
{
  return ang_pos;
}

void RBody::SetAngPos(Quaternion q)
{
  ang_pos = q;
  // normalizuje kwaternion by przywrocic mu dlugosc jednostkowa
  ang_pos = ang_pos.Normalize();
  
  rot_matrix_valid = false;
}

void RBody::RememberColPoint(Vector3 p)
{
  #ifdef COL_POINT_BUFFER
    recent_cols[recent_cols_last] = p;
    recent_cols_last = (recent_cols_last+1)%COL_BUF_SIZE;
    if (recent_cols_size < COL_BUF_SIZE)
      recent_cols_size++;
  #else
    return;
  #endif
}

void RBody::PushState()
{
  past_states.push_back(State());
  if (past_states.size() > STATE_LIST_LEN)
    past_states.pop_front();
}

void RBody::RevertState(int backward_count)
{
  if (backward_count <= 0)
    return;
    
  RBState old_state;
  for (int i = 0; i < backward_count-1; i++)
    past_states.pop_back();
    
  ApplyState(past_states.back());
  past_states.pop_back();
}
