#include "Resting.hpp"
#include "ODESolver.hpp"
#include "CDetector.hpp"
#include "CSolver.hpp"
#include "main.h"
#include <vector>
#include <list>
#include <stack>
#include <cstdio>
//#include "QuadProg/Array.hh"
//#include "QuadProg/QuadProg++.hh"

#define SEPARATION_FRAC 0.01

struct RestingCollision;

struct RestingBody
{
  RBody* body;
  int level;
  std::vector<RestingCollision> collisions;
};

struct RestingCollision
{
  Collision* collision;
  RestingBody* body;
};

void Separation(Collision* collision)
{
  RBody* A = collision->bodies[0];
  RBody* B = collision->bodies[1];
  
  float max_depth = -1;
  Vector3 max_normal;
  
  for (int body = 0; body < 2; body++)
  {
    for (unsigned int i = 0; i < collision->points[body].size(); i++)
    {
      CollisionPoint cp = collision->points[body][i];
      PointEvaluation pe = CDetector::EvaluatePoint(cp, collision->bodies[body], collision->bodies[(body+1)%2]);
      float rel_vel = CDetector::RelativeVelocity(pe.p, pe.n, collision->bodies[body],  collision->bodies[(body+1)%2]);
      if (pe.depth > max_depth && rel_vel < -0.001)
      {
        max_depth = pe.depth;
        max_normal = pe.n;
        if (body == 1)
          max_normal = max_normal * -1;
      }
    }
  }

  if (max_depth <= 0)
    return;
    
  if ((A->is_static || A->is_frozen) && (B->is_static || B->is_frozen))
    return;
  else if (A->is_static || A->is_frozen)
    B->pos -= max_normal*max_depth*SEPARATION_FRAC;
  else if (B->is_static || B->is_frozen)
    A->pos += max_normal*max_depth*SEPARATION_FRAC;
  else
  {
    float mass_sum = A->mass + B->mass;
    A->pos += max_normal*max_depth*(A->mass/mass_sum)*SEPARATION_FRAC;
    B->pos -= max_normal*max_depth*(B->mass/mass_sum)*SEPARATION_FRAC;
  }
}

void Resting::SolveAll(std::vector<RBody*>* bodies, double delta_time)
{  
  std::list<RestingBody*>   body_queue;
  std::vector<RestingBody>  body_vector;
  std::vector<RBState>      old_state;
  
  for (unsigned int i = 0; i < bodies->size(); i++)
  {
    RestingBody resting_body;
    resting_body.body = (*bodies)[i];
    resting_body.level = -2;
    body_vector.push_back(resting_body);
  }
  
  // find all collisions
  for (unsigned int i = 0; i < body_vector.size(); i++)
  {
    for (unsigned int j = 0; j < body_vector.size(); j++)
    {
      if (j == i) continue;
      RBState old_state = body_vector[i].body->State();
      ODESolver::EulerStepPos(*(body_vector[i].body), delta_time);
      if (CDetector::DetectCollision(body_vector[i].body, body_vector[j].body))
      {
        Collision* col = CDetector::GetCollisionPoints(body_vector[i].body, body_vector[j].body);
        if (col->points[0].size() == 0 && col->points[1].size() == 0) 
          continue;
        
        RestingCollision rest_col_i;
        rest_col_i.body = &(body_vector[j]);
        rest_col_i.collision = col;
        
        RestingCollision rest_col_j;
        rest_col_j.body = &(body_vector[i]);
        rest_col_j.collision = col;
        
        body_vector[i].collisions.push_back(rest_col_i);
        body_vector[j].collisions.push_back(rest_col_j);
      }
      body_vector[i].body->ApplyState(old_state);
    }
  }
  
  // fill queue with static bodies
  for (unsigned int i = 0; i < body_vector.size(); i++)
    if (body_vector[i].body->is_static)
    {
      body_vector[i].level = -1;
      body_queue.push_back(&(body_vector[i]));
    }
  
  // process the queue
  int max_level = -1;
  while (body_queue.size() > 0)
  {
    RestingBody* body = body_queue.front();
    body_queue.pop_front();
    if (body->level > max_level) max_level = body->level;
    
    for (unsigned int i = 0; i < body->collisions.size(); i++)
    {
      RestingBody* colliding_body = body->collisions[i].body;
      if (colliding_body->level == -2)
      {
        colliding_body->level = body->level + 1;
        body_queue.push_back(colliding_body);
      }
    }
  }
  
  std::vector<RestingBody*>* body_levels = new std::vector<RestingBody*>[max_level+1];
  std::vector<Collision*>* collision_levels = new std::vector<Collision*>[max_level+1];
  
  for (unsigned int i = 0; i < body_vector.size(); i++)
  {
    RestingBody* body_a = &(body_vector[i]);
    if (body_a->level < 0) continue;
    
    body_levels[body_a->level].push_back(body_a);
    for (unsigned int j = 0; j < body_a->collisions.size(); j++)
    { 
      RestingBody* body_b = body_a->collisions[j].body;
      if (body_b->level <= body_a->level)
      {
        bool found = false;
        for (unsigned int ci = 0; ci < collision_levels[body_a->level].size(); ci++)
        {
          if (collision_levels[body_a->level][ci] == body_a->collisions[j].collision)
            found = true;
        }
        if (!found)
          collision_levels[body_a->level].push_back(body_a->collisions[j].collision);
      }
    }
  }
  global::DumpInfo();
  // RESTING
  for (int iter = 1; iter >= 0; iter--)
  {
    double elasticity = -double(iter)*0.1f;
    for (int lvl = 0; lvl < max_level+1; lvl++)
    {
      for (unsigned int i = 0; i < collision_levels[lvl].size(); i++)
      {
        Collision* col = new Collision(*(collision_levels[lvl][i]));
        
        global::DumpLevels(lvl, col->bodies[0]->id, col->bodies[1]->id);
        //if (iter == 0)
        //printf("level%i: %s, %s: %i, %i\n", lvl, col->bodies[0]->name.c_str(), col->bodies[1]->name.c_str(), col->points[0].size(), col->points[1].size());
        
        CSolver::SolveContact(col, false, elasticity);
        delete col;
      }
    }
  }
  global::DumpInfo();
  // SHOCK PROPAGATION
  for (int iter = 0; iter >= 0; iter--)
  {
    for (int lvl = 0; lvl < max_level+1; lvl++)
    {
      for (unsigned int i = 0; i < collision_levels[lvl].size(); i++)
      {
        Collision* col = new Collision(*(collision_levels[lvl][i]));
        CSolver::SolveContact(col, false, 0);
        delete col;
      }
      for (unsigned int i = 0; i < body_levels[lvl].size(); i++)
        body_levels[lvl][i]->body->is_frozen = true;
    }
    
    for (int lvl = 0; lvl < max_level+1; lvl++)
    {
      for (unsigned int i = 0; i < body_levels[lvl].size(); i++)
        body_levels[lvl][i]->body->is_frozen = false;
    }
  }
  
  // OFFSETTING
  for (int iter = 0; iter < 2; iter++)
  {
    for (int lvl = 0; lvl < max_level+1; lvl++)
    {
      for (unsigned int i = 0; i < collision_levels[lvl].size(); i++)
      {
        Collision* col = new Collision(*(collision_levels[lvl][i]));
        Separation(col);
        delete col;
      }
      for (unsigned int i = 0; i < body_levels[lvl].size(); i++)
        body_levels[lvl][i]->body->is_frozen = true;
    }
    
    for (int lvl = 0; lvl < max_level+1; lvl++)
    {
      for (unsigned int i = 0; i < body_levels[lvl].size(); i++)
        body_levels[lvl][i]->body->is_frozen = false;
    }
  }
}

/*
double compute_aij(Collision* ci, CollisionPoint& cpi, Collision* cj, CollisionPoint& cpj)
{
  // If the bodies involved in the ith and jth contact are
  // distinct, then a ij is zero.
  if((ci->A != cj->A) && (ci->B != cj->B) && (ci->A != cj->B) && (ci->B != cj->A))
    return 0.0;
    
  RBody* A = ci->A;
  RBody* B = cj->B;
  
  Vector3 ni = cpi.n;
  Vector3 nj = cpj.n;
  Vector3 pi = cpi.p;
  Vector3 pj = cpj.p;
  Vector3 ra = pi - A->pos;
  Vector3 rb = pi - B->pos;
  
  // What force and torque does contact j exert on body A?
  Vector3 force_on_a;
  Vector3 torque_on_a;
  if (cj->A == ci->A)
  {
    force_on_a = nj;
    torque_on_a = (pj - A->pos) ^ nj;
  }
  else if(cj->B == ci->A)
  {
    force_on_a = nj * -1;
    torque_on_a = (pj - A->pos) ^ nj;
  }
  
  // What force and torque does contact j exert on body B?
  Vector3 force_on_b;
  Vector3 torque_on_b;
  if(cj->A == ci->B)
  {
    force_on_b = nj;
    torque_on_b = (pj - B->pos) ^ nj;
  }
  else if(cj->B == ci->B)
  {
    force_on_b = nj * -1;
    torque_on_b = (pj - B->pos) ^ nj;
  }

  // Now compute how the jth contact force affects the linear
  // and angular acceleration of the contact point on body A
  Vector3 a_linear = force_on_a / A->mass;
  Vector3 a_angular = (A->InvInertiaTensor() * torque_on_a) ^ ra;
  // Same for B
  Vector3 b_linear = force_on_b / B->mass;
  Vector3 b_angular = (B->InvInertiaTensor() * torque_on_b) ^ rb;
  
  return ni * ((a_linear + a_angular) - (b_linear + b_angular));
}

Matrix<double> ComputeMatrixA(std::vector<Collision*>* collisions, int point_num)
{
  Matrix<double> A;
  A.resize(point_num, point_num);
  
  int i = 0;
  int j = 0;
  
  for (int ic = 0; ic < point_num; ic++)
  {
    for (int ip = 0; ip < point_num; ip++)
    {
      for (int jc = 0; jc < point_num; jc++)
      {
        for (int jp = 0; jp < point_num; jp++)
        {
          Collision* collision_i = (*collisions)[ic];
          CollisionPoint cpoint_i = collision_i->points[ip];
          Collision* collision_j = (*collisions)[jc];
          CollisionPoint cpoint_j = collision_i->points[jp];
          A[i][j] = compute_aij(collision_i, cpoint_i, collision_j, cpoint_j);
          j++;
        }
      }
      i++;
    }
  }
  
  return A;
}

Vector<double> ComputeVectorB(std::vector<Collision*>* collisions, int point_num)
{
  Vector<double> B;
  B.resize(point_num);
  
  for (unsigned int i = 0; i < collisions->size(); i++)
  {
    for (unsigned int j = 0; j < (*collisions)[i]->points.size(); j++)
    {
      Collision *col = (*collisions)[i];
      CollisionPoint deepest = col->points[j];
      
      RBody* Body_A = col->A;
      RBody* Body_B = col->B;
      
      Vector3 n = deepest.n;
      Vector3 p = deepest.p;
      Vector3 ra = p - Body_A->pos;
      Vector3 rb = p - Body_B->pos;
      
      // siły zewnętrzne
      Vector3 FA = Body_A->Forces(0, Body_A->State());
      Vector3 FB = Body_A->Forces(0, Body_B->State());
      Vector3 TA = Body_A->Torque(0, Body_A->State());
      Vector3 TB = Body_B->Torque(0, Body_B->State());
      
      // Elementy p bis A/B niezależne od f, najpierw zależne od sił zewnętrznych
      // a następnie od prędkości

      Vector3 a_ext_part = FA / Body_A->mass + ((Body_A->InvInertiaTensor() * TA) ^ ra);
      Vector3 b_ext_part = FB / Body_B->mass + ((Body_B->InvInertiaTensor() * TB) ^ rb);
      
      Vector3 a_vel_part = (Body_A->ang_vel ^ (Body_A->ang_vel ^ ra)) + 
                  ((Body_A->InvInertiaTensor() * ((Body_A->InertiaTensor()*Body_A->ang_vel) ^ Body_A->ang_vel)) ^ ra);
      Vector3 b_vel_part = (Body_B->ang_vel ^ (Body_B->ang_vel ^ rb)) + 
                  ((Body_B->InvInertiaTensor() * ((Body_B->InertiaTensor()*Body_B->ang_vel) ^ Body_B->ang_vel)) ^ rb);
      
      // ONLY FOR VERTEX-FACE COLLISION
      Vector3 nprim = Body_B->ang_vel ^ n;
      Vector3 pAvel = Body_A->vel + (Body_A->ang_vel ^ ra);
      Vector3 pBvel = Body_B->vel + (Body_B->ang_vel ^ rb);
      
      double k1 = n * ((a_ext_part + a_vel_part) - (b_ext_part + b_vel_part));
      double k2 = (nprim*2) * (pAvel - pBvel);
      B[i] = k1 + k2;
    }
  }
  
  return B;
}

void Resting::ApplyContactForces(std::vector<Collision*>* contacts)
{
  int size = 0;
  for (unsigned int i = 0; i < contacts->size(); i++)
    size += (*contacts)[i]->points.size();
  
  Matrix<double> A = ComputeMatrixA(contacts, size);
  Vector<double> b = ComputeVectorB(contacts, size);
  
  Matrix<double> CE, CI;
  Vector<double> ce0, ci0, x;
  
  CI.resize(size, 2*size);
  for (int i = 0; i < size; i++)
  {
    for (int j = 0; j < size; j++)
    {
      if (i == j) CI[i][j] = 1;
      else CI[i][j] = 0;
    }
    for (int j = size; j < 2*size; j++)
    {
      CI[i][j] = A[i][j];
    }
  }
    
  ci0.resize(2*size);
  for (int i = 0; i < size; i++)
    ci0[i] = 0;
  for (int i = size; i < 2*size; i++)
    ci0[i] = b[i];
    
  CE.resize(size,1);
  for (int i = 0; i < size; i++)
    CE[i][0] = 0;
    
  ce0.resize(1);
  ce0[0] = 0;
  
  x.resize(size);
  
  std::cout << "CE: " << CE << std::endl;
  std::cout << "ce0: " << ce0 << std::endl;
  std::cout << "CI: " << CI << std::endl;
  std::cout << "ci0: " << ci0 << std::endl;
  
  A *= 2.0f;
  std::cout << "f: " << solve_quadprog(A, b, CE, ce0, CI, ci0, x) << std::endl;
  
  int xi = 0;
  for (unsigned int i = 0; i < contacts->size(); i++)
  {
    Collision* contact = (*contacts)[i];
    for (unsigned int j = 0; j < contact->points.size(); j++)
    {
      RBody* Body_A = contact->A;
      RBody* Body_B = contact->B;
      Vector3 n = contact->points[j].n;
      Vector3 p = contact->points[j].p;
      
      Body_A->forces += n * x[xi];
      Body_A->torque += (n * x[xi]) ^ (p - Body_A->pos);
      Body_B->forces -= n * x[xi];
      Body_B->torque -= (n * x[xi]) ^ (p - Body_B->pos);
    }
  }
  
  //return x;
}
*/
