#include "RBody.hpp"
#include "CDetector.hpp"
#include "ODESolver.hpp"
#include "CSolver.hpp"
#include "Renderer.hpp"
#include "Resting.hpp"
#include "Network.hpp"
#include "Player.hpp"
#include "main.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <vector>
#include <unistd.h>
#include <GL/glut.h>
#include <fstream>
#include <cstring>

#define EPS 0
#define STEP_TIME 0.04
#define MAX_STEPS_PER_FRAME 10

namespace global
{
  double speed = 1.0f;
  int frames = 0;
  float fps = 0;
  double fps_timer = 0;
  double GLUT_time;
  double camera_angle;
  double time_acc;
  
  bool old_keys[256];
  bool new_keys[256];
  
  bool mouse_key;
  bool scroll_down;
  bool scroll_up;
  int mouse_x_old;
  int mouse_x_new;
  int mouse_y_old;
  int mouse_y_new;
  
  Quaternion camera_rotation;
  
  unsigned int simulation_time;
  
  bool is_host;
  
  //benchmarking
  double frame_time_acc = 0;
  int frame_num_total = 0;
  
  std::vector<RBody*> bodies;
  std::vector<RBState> old_state;
  std::vector<Collision*> collisions;
  std::vector<Collision*> contacts;
  std::vector<Player> players;
  unsigned int player_id;

  std::ofstream dumpfile;
  
  void DumpInfo()
  {
    if (!dumpfile.is_open())
    {
      std::string filename;
      filename  = "dump";
      filename += '0'+player_id;
      filename += ".txt";
      dumpfile.open(filename);
    }
    
    dumpfile << "t: " << simulation_time << "\n";
    for (unsigned int i = 0; i < players.size(); i++)
    {
      dumpfile << " p" << i << "input: ";
      dumpfile << players[i].forward << "," << players[i].backward << ",";
      dumpfile << players[i].left << "," << players[i].right << "\n";
    }
    for (unsigned int i = 0; i < bodies.size(); i++)
    {
      dumpfile << " body" << i << "\n";
      dumpfile << "  pos     " << bodies[i]->pos[0] << "," << bodies[i]->pos[1] << "," << bodies[i]->pos[2] << "\n";
      dumpfile << "  vel     " << bodies[i]->vel[0] << "," << bodies[i]->vel[1] << "," << bodies[i]->vel[2] << "\n";
      dumpfile << "  ang_vel " << bodies[i]->ang_vel[0] << "," << bodies[i]->ang_vel[1] << "," << bodies[i]->ang_vel[2] << "\n";
      dumpfile << "  ang_pos " 
        << bodies[i]->GetAngPos().s << "," 
        << bodies[i]->GetAngPos().v[0] << "," 
        << bodies[i]->GetAngPos().v[1] << "," 
        << bodies[i]->GetAngPos().v[2] << "\n";
    }
    dumpfile << "\n";
  }
  
  void DumpLevels(int lvl, int a, int b)
  {
    dumpfile << " lvl" << lvl << " " << a << " " << b << "\n";
  }
  
  void DumpPoints(Vector3 max_point, double depth)
  {
    dumpfile << " mpoint: " << max_point[0] << "," << max_point[1] << "," << max_point[2] << "\n";
    dumpfile << " " << depth << "\n";
  }
}

using namespace global;

void UpdateControls()
{
  if (old_keys['w'] != new_keys['w'])
  {
    players[player_id].forward = new_keys['w'];
    Network::SendInput(players[player_id].forward ? Network::FORWARD_DOWN : Network::FORWARD_UP, simulation_time, player_id);
  }
  if (old_keys['s'] != new_keys['s'])
  {
    players[player_id].backward = new_keys['s'];
    Network::SendInput(players[player_id].backward ? Network::BACKWARD_DOWN : Network::BACKWARD_UP, simulation_time, player_id);
  }
  if (old_keys['a'] != new_keys['a'])
  {
    players[player_id].left = new_keys['a'];
    Network::SendInput(players[player_id].left ? Network::LEFT_DOWN : Network::LEFT_UP, simulation_time, player_id);
  }
  if (old_keys['d'] != new_keys['d'])
  {
    players[player_id].right = new_keys['d'];
    Network::SendInput(players[player_id].right ? Network::RIGHT_DOWN : Network::RIGHT_UP, simulation_time, player_id);
  }
  
  for (int i = 0; i < 256; i++)
    old_keys[i] = new_keys[i];
}

void RotateCamera()
{
  Renderer::RotateCamera(mouse_y_new - mouse_y_old, mouse_x_new - mouse_x_old, scroll_up, scroll_down);
  
  mouse_x_old = mouse_x_new;
  mouse_y_old = mouse_y_new;
  scroll_down = false;
  scroll_up = false;
}

void ControlBody(Player player)
{
  if (player.controlled_body == NULL)
    return;
  
  double force = 20*player.controlled_body->mass;
    
  if (player.forward)  player.controlled_body->forces -= Vector3(    0, force, 0);
  if (player.backward) player.controlled_body->forces += Vector3(    0, force, 0);
  if (player.right)    player.controlled_body->forces -= Vector3(force,     0, 0);
  if (player.left)     player.controlled_body->forces += Vector3(force,     0, 0);
}

void ResetForces()
{
  for (unsigned int i = 0; i < bodies.size(); i++)
  {
    bodies[i]->forces = Vector3(0, 0, -(bodies[i]->mass)*9.8f);
  }
}

void PushStates()
{
  for (unsigned int i = 0; i < bodies.size(); i++)
  {
    bodies[i]->PushState();
  }
}

void SimulationStep(double delta_time)
{  
  if (bodies.size() == 0)
    return;
  
  ResetForces();
  for (unsigned int i = 0; i < players.size(); i++)
    ControlBody(players[i]);
  PushStates();
  
  // COLLISION PROCESSING
  for (int i = 0; i < 1; i++)
  {
    for (unsigned int i = 0; i < bodies.size(); i++)
      old_state[i] = bodies[i]->State();
    
    ODESolver::EulerStepAll(&bodies, delta_time);
    CDetector::CheckAll(&bodies, &collisions, &contacts, EPS);
    //printf("col: %d, cont: %d\n", collisions.size(), contacts.size());
    
    for (unsigned int i = 0; i < bodies.size(); i++)
      bodies[i]->ApplyVel(old_state[i]);
  
    for (int j = 0; j < 3; j++)
      CSolver::SolveAllCollisions(&collisions, delta_time);
    
    for (unsigned int i = 0; i < bodies.size(); i++)
      bodies[i]->ApplyPos(old_state[i]);
  }

  ODESolver::EulerStepAllVel(&bodies, delta_time);
  for (unsigned int i = 0; i < bodies.size(); i++)
  {
    bodies[i]->ResetForceTorque();
  }

  Resting::SolveAll(&bodies, delta_time);
  ODESolver::EulerStepAllPos(&bodies, delta_time);
  
  time_acc -= delta_time;
  simulation_time += 1;
}

void HandleCommand(Network::Command command)
{
  if (command.type == Network::MSG_SYNC_BODIES)
  {
    simulation_time = command.simulation_time;
    player_id = command.id;
    for (unsigned int i = 0; i <= player_id; i++)
    {
      players.push_back(Player(i, bodies[i]));
    }
  }
  else if (command.type == Network::MSG_SEND_INPUT)
  {
    //printf("\ttime: %i, command: %i\n", simulation_time, command.simulation_time);
    while (simulation_time < command.simulation_time)
    {
      SimulationStep(STEP_TIME);
    }
    
    unsigned int time_diff = (simulation_time - command.simulation_time);
    
    switch (command.input)
    {
      case Network::FORWARD_DOWN:
        players[command.id].forward = true;
        break;
      case Network::FORWARD_UP:
        players[command.id].forward = false;
        break;
      case Network::BACKWARD_DOWN:
        players[command.id].backward = true;
        break;
      case Network::BACKWARD_UP:
        players[command.id].backward = false;
        break;
      case Network::LEFT_DOWN:
        players[command.id].left = true;
        break;
      case Network::LEFT_UP:
        players[command.id].left = false;
        break;
      case Network::RIGHT_DOWN:
        players[command.id].right = true;
        break;
      case Network::RIGHT_UP:
        players[command.id].right = false;
        break;
    }
    
    //printf("\tReverting %i, timer: %i, input: %i\n", time_diff, simulation_time, command.input);
    
    for (unsigned int i = 0; i < bodies.size(); i++)
    {
      bodies[i]->RevertState(time_diff);
    }
    simulation_time = command.simulation_time;
    
    //printf("\tFast forward.\n");
    
    for (unsigned int i = 0; i < time_diff; i++)
    {
      SimulationStep(STEP_TIME);
    }
    
    //printf("\tDone, timer: %i\n", simulation_time);
  }
}

static void Idle(void)
{
  double new_GLUT_time = ((double)glutGet(GLUT_ELAPSED_TIME)) / 1000.0f;
  double single_frame_time = ((double)glutGet(GLUT_ELAPSED_TIME)) / 1000.0f;
  
  double time_elapsed = new_GLUT_time - GLUT_time;
  fps_timer += time_elapsed;
  frames++;
  if (fps_timer >= 1.0f)
  {
    //printf("%f\n", fps);
    fps = ((double)frames)/fps_timer;
    frames = 0;
    fps_timer = 0;
  }
  
  double delta_time = STEP_TIME*speed;
  time_acc += time_elapsed*speed;
  
  UpdateControls();
  
  if (is_host)
    HandleCommand(Network::HostLoop(simulation_time, &speed));
  else
    HandleCommand(Network::ClientLoop(&speed));
  
  if (old_state.size() != bodies.size())
  {
    old_state.clear();
    for (unsigned int i = 0; i < bodies.size(); i++)
    {
      old_state.push_back(bodies[i]->State());
    }
  }
  
  int step_cnt = 0;
  while (time_acc >= delta_time && speed > 0 && step_cnt < MAX_STEPS_PER_FRAME)
  {
    SimulationStep(delta_time);
    step_cnt++;
  }  
  /*while (collisions.size() > 0)
  {
    double search_step = delta_time/2.0f;
    double max_depth = CDetector::MaxPen(&collisions);

    while (max_depth > EPS || max_depth < -EPS)
    {
      //usleep(100000);
      if (max_depth > EPS) delta_time = -search_step;
      else delta_time = search_step;
      search_step /= 2.0f;
      solved_time += delta_time;
      ODESolver::RK4StepAll(&bodies, solved_time, delta_time);
      CDetector::RecheckAll(&collisions);
      max_depth = CDetector::MaxPen(&collisions);
    }
    CSolver::SolveAll(&collisions, EPS);
    delta_time = simulation_time - solved_time;
    ODESolver::RK4StepAll(&bodies, solved_time, delta_time);
    CDetector::RecheckAll(&collisions);
  }*/
  
  //speed = 0;
  
  GLUT_time = new_GLUT_time;
  frame_time_acc += ((double)glutGet(GLUT_ELAPSED_TIME)) / 1000.0f - single_frame_time;
  frame_num_total++;
  
  if (frame_num_total >= 100)
  {
    //printf("avg over 100 frames: %f\n", frame_time_acc / frame_num_total);
    //printf("%i, %s\n", bodies[0]->is_frozen, bodies[0]->name.c_str());
    frame_num_total = 0;
    frame_time_acc = 0;
  }
  
  RotateCamera();
  glutPostRedisplay();
}

int PrepareScene(const char* filename)
{
  bodies.clear();
  
  std::ifstream file(filename);
  if (!file.is_open())
  {
    printf("Scene file not found!\n");
    return -1;
  }
  
  file.ignore(1024,'~');
  
  double a, b, c, d;
  std::string name;
  bool is_static;
  int last_id = 0;
  
  while(file.peek() != ';')
  {
    file >> a >> b >> c >> d >> name;
    RBody* body = new RBody(a,b,c,d,name);
    body->id = last_id;
    last_id++;

    file >> is_static;
    body->is_static = is_static;
    
    file >> a >> b >> c;
    body->pos = Vector3(a,b,c);
    
    if (!is_static)
    {
      file >> a >> b >> c;
      body->vel = Vector3(a,b,c);
    }
    
    file >> a >> b >> c >> d;
    body->SetAngPos(Quaternion(a,b,c,d).Normalize());

    if (!is_static)
    {
      file >> a >> b >> c;
      body->ang_vel = Vector3(a,b,c);
    }
    
    bodies.push_back(body);
  }
  
  file.close();
  
  old_state.clear();
  for (unsigned int i = 0; i < bodies.size(); i++)
  {
    old_state.push_back(bodies[i]->State());
  }
  
  return 0;
}

int Init(int* argc, char** argv)
{
  Renderer::InitDisplay(argc, argv, &bodies, &collisions, 800, 600);
  
  if (*argc < 3)
    return -1;
    
  if (strcmp(argv[1], "h") == 0)
  {
    if (PrepareScene(argv[2]) != 0)
      return -1;
    Network::CreateHost(&bodies, &players);
    is_host = true;
    player_id = 0;
    players.push_back(Player(player_id, bodies[player_id]));
  }
  else if (strcmp(argv[1], "c") == 0)
  {
    Network::Connect(argv[2], &bodies, &players);
    is_host = false;
  }
  else
  {
    printf("Unrecognized option, first argument must be 'h' for host or 'c' for client.\n");
    return -1;
  }
  
  return 0;
}

void Keyboard(unsigned char key, int x, int y)
{
  new_keys[key] = true;
  
  if (key == 'q' || key == 'Q') exit(0);
  if (key == ' ')
  {
    if (speed == 1) speed = 0;
    else if (speed == 0) speed = 1;
  }
  if (key == '+') Renderer::SetCamera(camera_angle += 0.1);
  if (key == '-') Renderer::SetCamera(camera_angle -= 0.1);
}

void KeyboardUp(unsigned char key, int x, int y)
{
  new_keys[key] = false;
}

void Mouse(int button, int state, int x, int y) 
{
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) 
  {
    mouse_key = true;
    mouse_x_old = mouse_x_new = x;
    mouse_y_old = mouse_y_new = y;
  }
  else if (button == 3 && state == GLUT_DOWN) 
  {
    scroll_up = true;
  }
  else if (button == 4 && state == GLUT_DOWN) 
  {
    scroll_down = true;
  }
  else
  {
    mouse_key = false;
  }
}
 
void MouseMotion(int x, int y)
{
  if (mouse_key) 
  {
    mouse_x_new = x;
    mouse_y_new = y;
  }
}

int main(int argc, char** argv)
{
  if (Init(&argc, argv) != 0)
  {
    printf("Failed to initialise!\n");
    return -1;
  }
  
  glutSetKeyRepeat(GLUT_KEY_REPEAT_OFF);
  glutKeyboardFunc(Keyboard);
  glutKeyboardUpFunc(KeyboardUp);
  glutMouseFunc(Mouse);
  glutMotionFunc(MouseMotion);
  glutIdleFunc(Idle);
  glutMainLoop();

  dumpfile.close();

  return 0;
}
